# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import math
import os.path
import re
import textwrap
from collections import namedtuple
from unittest import mock

import pytest
from pytest import approx

from cloudinit.config import cc_mounts
from cloudinit.config.cc_mounts import (
    GB,
    MB,
    create_swapfile,
    suggested_swapsize,
)
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.subp import ProcessExecutionError
from tests.unittests import helpers as test_helpers

M_PATH = "cloudinit.config.cc_mounts."


class TestSanitizeDevname:
    def _touch(self, path, new_root):
        path = os.path.join(new_root, path.lstrip("/"))
        basedir = os.path.dirname(path)
        if not os.path.exists(basedir):
            os.makedirs(basedir)
        open(path, "a").close()

    def _makedirs(self, directory, new_root):
        directory = os.path.join(new_root, directory.lstrip("/"))
        if not os.path.exists(directory):
            os.makedirs(directory)

    def mock_existence_of_disk(self, disk_path, new_root):
        self._touch(disk_path, new_root)
        self._makedirs(
            os.path.join("/sys/block", disk_path.split("/")[-1]), new_root
        )

    def mock_existence_of_partition(
        self, disk_path, partition_number, new_root
    ):
        self.mock_existence_of_disk(disk_path, new_root)
        self._touch(disk_path + str(partition_number), new_root)
        disk_name = disk_path.split("/")[-1]
        self._makedirs(
            os.path.join(
                "/sys/block", disk_name, disk_name + str(partition_number)
            ),
            new_root,
        )

    def test_existent_full_disk_path_is_returned(self, fake_filesystem):
        disk_path = "/dev/sda"
        self.mock_existence_of_disk(disk_path, fake_filesystem)
        assert disk_path == cc_mounts.sanitize_devname(
            disk_path, lambda x: None
        )

    def test_existent_disk_name_returns_full_path(self, fake_filesystem):
        disk_name = "sda"
        disk_path = "/dev/" + disk_name
        self.mock_existence_of_disk(disk_path, fake_filesystem)
        assert disk_path == cc_mounts.sanitize_devname(
            disk_name, lambda x: None
        )

    def test_existent_meta_disk_is_returned(self, fake_filesystem):
        actual_disk_path = "/dev/sda"
        self.mock_existence_of_disk(actual_disk_path, fake_filesystem)
        assert actual_disk_path == cc_mounts.sanitize_devname(
            "ephemeral0",
            lambda x: actual_disk_path,
        )

    def test_existent_meta_partition_is_returned(self, fake_filesystem):
        disk_name, partition_part = "/dev/sda", "1"
        actual_partition_path = disk_name + partition_part
        self.mock_existence_of_partition(
            disk_name, partition_part, fake_filesystem
        )
        assert actual_partition_path == cc_mounts.sanitize_devname(
            "ephemeral0.1",
            lambda x: disk_name,
        )

    def test_existent_meta_partition_with_p_is_returned(self, fake_filesystem):
        disk_name, partition_part = "/dev/sda", "p1"
        actual_partition_path = disk_name + partition_part
        self.mock_existence_of_partition(
            disk_name, partition_part, fake_filesystem
        )
        assert actual_partition_path == cc_mounts.sanitize_devname(
            "ephemeral0.1",
            lambda x: disk_name,
        )

    def test_first_partition_returned_if_existent_disk_is_partitioned(
        self, fake_filesystem
    ):
        disk_name, partition_part = "/dev/sda", "1"
        actual_partition_path = disk_name + partition_part
        self.mock_existence_of_partition(
            disk_name, partition_part, fake_filesystem
        )
        assert actual_partition_path == cc_mounts.sanitize_devname(
            "ephemeral0",
            lambda x: disk_name,
        )

    def test_nth_partition_returned_if_requested(self, fake_filesystem):
        disk_name, partition_part = "/dev/sda", "3"
        actual_partition_path = disk_name + partition_part
        self.mock_existence_of_partition(
            disk_name, partition_part, fake_filesystem
        )
        assert actual_partition_path == cc_mounts.sanitize_devname(
            "ephemeral0.3",
            lambda x: disk_name,
        )

    def test_transformer_returning_none_returns_none(self, fake_filesystem):
        assert (
            cc_mounts.sanitize_devname(
                "ephemeral0",
                lambda x: None,
            )
            is None
        )

    def test_missing_device_returns_none(self, fake_filesystem):
        assert (
            cc_mounts.sanitize_devname(
                "/dev/sda",
                None,
            )
            is None
        )

    def test_missing_sys_returns_none(self, fake_filesystem):
        disk_path = "/dev/sda"
        self._makedirs(disk_path, fake_filesystem)
        assert (
            cc_mounts.sanitize_devname(
                disk_path,
                None,
            )
            is None
        )

    def test_existent_disk_but_missing_partition_returns_none(
        self, fake_filesystem
    ):
        disk_path = "/dev/sda"
        self.mock_existence_of_disk(disk_path, fake_filesystem)
        assert (
            cc_mounts.sanitize_devname(
                "ephemeral0.1",
                lambda x: disk_path,
            )
            is None
        )

    def test_network_device_returns_network_device(self, fake_filesystem):
        disk_path = "netdevice:/path"
        assert disk_path == cc_mounts.sanitize_devname(
            disk_path,
            None,
        )

    def test_device_aliases_remapping(self, fake_filesystem):
        disk_path = "/dev/sda"
        self.mock_existence_of_disk(disk_path, fake_filesystem)
        assert disk_path == cc_mounts.sanitize_devname(
            "mydata", lambda x: None, {"mydata": disk_path}
        )


class TestSwapFileCreation:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, fake_filesystem: str):
        self.new_root = fake_filesystem
        self.swap_path = os.path.join(fake_filesystem, "swap.img")
        fstab_path = os.path.join(fake_filesystem, "etc/fstab")
        self._makedirs("/etc")

        self.m_fstab = mocker.patch(f"{M_PATH}FSTAB_PATH", fstab_path)
        self.m_subp = mocker.patch(f"{M_PATH}subp.subp")
        self.m_mounts = mocker.patch(
            f"{M_PATH}util.mounts",
            return_value={
                "/dev/sda1": {
                    "fstype": "ext4",
                    "mountpoint": "/",
                    "opts": "rw,relatime,discard",
                }
            },
        )

        self.mock_cloud = mock.Mock()
        self.mock_log = mock.Mock()
        self.mock_cloud.device_name_to_device = self.device_name_to_device

        self.cc = {
            "swap": {
                "filename": self.swap_path,
                "size": "512",
                "maxsize": "512",
            }
        }

    def _makedirs(self, directory):
        directory = os.path.join(self.new_root, directory.lstrip("/"))
        if not os.path.exists(directory):
            os.makedirs(directory)

    def device_name_to_device(self, path):
        if path == "swap":
            return self.swap_path
        else:
            dev = None

        return dev

    @mock.patch("cloudinit.util.get_mount_info")
    @mock.patch("cloudinit.util.kernel_version")
    def test_swap_creation_method_fallocate_on_xfs(
        self, m_kernel_version, m_get_mount_info
    ):
        m_kernel_version.return_value = (4, 20)
        m_get_mount_info.return_value = ["", "xfs"]

        cc_mounts.handle("", self.cc, self.mock_cloud, [])
        self.m_subp.assert_has_calls(
            [
                mock.call(
                    ["fallocate", "-l", "0M", self.swap_path], capture=True
                ),
                mock.call(["mkswap", self.swap_path]),
                mock.call(["swapon", "-a"]),
            ]
        )

    @mock.patch("cloudinit.util.get_mount_info")
    @mock.patch("cloudinit.util.kernel_version")
    def test_swap_creation_method_xfs(
        self, m_kernel_version, m_get_mount_info
    ):
        m_kernel_version.return_value = (3, 18)
        m_get_mount_info.return_value = ["", "xfs"]

        cc_mounts.handle("", self.cc, self.mock_cloud, [])
        self.m_subp.assert_has_calls(
            [
                mock.call(
                    [
                        "dd",
                        "if=/dev/zero",
                        "of=" + self.swap_path,
                        "bs=1M",
                        "count=0",
                    ],
                    capture=True,
                ),
                mock.call(["mkswap", self.swap_path]),
                mock.call(["swapon", "-a"]),
            ]
        )

    @mock.patch("cloudinit.util.get_mount_info")
    @mock.patch("cloudinit.util.kernel_version")
    def test_swap_creation_method_btrfs(
        self, m_kernel_version, m_get_mount_info
    ):
        m_kernel_version.return_value = (4, 20)
        m_get_mount_info.return_value = ["", "btrfs"]

        cc_mounts.handle("", self.cc, self.mock_cloud, [])
        self.m_subp.assert_has_calls(
            [
                mock.call(["truncate", "-s", "0", self.swap_path]),
                mock.call(["chattr", "+C", self.swap_path]),
                mock.call(
                    ["fallocate", "-l", "0M", self.swap_path],
                    capture=True,
                ),
                mock.call(["mkswap", self.swap_path]),
                mock.call(["swapon", "-a"]),
            ]
        )

    @mock.patch("cloudinit.util.get_mount_info")
    @mock.patch("cloudinit.util.kernel_version")
    def test_swap_creation_method_ext4(
        self, m_kernel_version, m_get_mount_info
    ):
        m_kernel_version.return_value = (5, 14)
        m_get_mount_info.return_value = ["", "ext4"]

        cc_mounts.handle("", self.cc, self.mock_cloud, [])
        self.m_subp.assert_has_calls(
            [
                mock.call(
                    ["fallocate", "-l", "0M", self.swap_path], capture=True
                ),
                mock.call(["mkswap", self.swap_path]),
                mock.call(["swapon", "-a"]),
            ]
        )


class TestFstabHandling:
    swap_path = "/dev/sdb1"

    @pytest.fixture(autouse=True)
    def setup(self, mocker, fake_filesystem: str):
        self.new_root = fake_filesystem

        self.fstab_path = os.path.join(self.new_root, "etc/fstab")
        self._makedirs("/etc")

        self.m_fstab = mocker.patch(f"{M_PATH}FSTAB_PATH", self.fstab_path)
        self.m_subp = mocker.patch(f"{M_PATH}subp.subp")
        self.m_mounts = mocker.patch(
            f"{M_PATH}util.mounts",
            return_value={
                "/dev/sda1": {
                    "fstype": "ext4",
                    "mountpoint": "/",
                    "opts": "rw,relatime,discard",
                }
            },
        )

        self.m_is_block_device = mocker.patch(
            f"{M_PATH}_is_block_device", return_value=True
        )

        self.mock_cloud = mock.Mock()
        self.mock_log = mock.Mock()
        self.mock_cloud.device_name_to_device = self.device_name_to_device

    def _makedirs(self, directory):
        directory = os.path.join(self.new_root, directory.lstrip("/"))
        if not os.path.exists(directory):
            os.makedirs(directory)

    def device_name_to_device(self, path):
        if path == "swap":
            return self.swap_path
        else:
            dev = None

        return dev

    def test_no_fstab(self):
        """Handle images which do not include an fstab."""
        assert not os.path.exists(cc_mounts.FSTAB_PATH)
        fstab_expected_content = (
            "%s\tnone\tswap\tsw,comment=cloudconfig\t0\t0\n"
            % (self.swap_path,)
        )
        cc_mounts.handle("", {}, self.mock_cloud, [])
        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    @pytest.mark.parametrize(
        "fstype, expected",
        [
            (
                "btrfs",
                [
                    mock.call(["truncate", "-s", "0", "/swap.img"]),
                    mock.call(["chattr", "+C", "/swap.img"]),
                    mock.call(
                        ["fallocate", "-l", "0M", "/swap.img"], capture=True
                    ),
                ],
            ),
            (
                "xfs",
                [
                    mock.call(
                        [
                            "dd",
                            "if=/dev/zero",
                            "of=/swap.img",
                            "bs=1M",
                            "count=0",
                        ],
                        capture=True,
                    )
                ],
            ),
            (
                "ext4",
                [
                    mock.call(
                        ["fallocate", "-l", "0M", "/swap.img"], capture=True
                    )
                ],
            ),
        ],
    )
    def test_swap_creation_command(self, fstype, expected, mocker):
        """Ensure that the swap file is correctly created.

        Different filesystems require different methods.
        """
        mocker.patch(
            "cloudinit.util.get_mount_info", return_value=["", fstype]
        )
        mocker.patch("cloudinit.util.kernel_version", return_value=(4, 17))

        fstab = "/swap.img swap swap defaults 0 0\n"

        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab)
        cc = {
            "swap": {"filename": "/swap.img", "size": "512", "maxsize": "512"}
        }
        cc_mounts.handle("", cc, self.mock_cloud, [])
        assert self.m_subp.call_args_list == expected + [
            mock.call(["mkswap", "/swap.img"]),
            mock.call(["swapon", "-a"]),
            mock.call(["mount", "-a"]),
            mock.call(["systemctl", "daemon-reload"]),
        ]

    def test_fstab_no_swap_device(self):
        """Ensure that cloud-init adds a discovered swap partition
        to /etc/fstab."""

        fstab_original_content = ""
        fstab_expected_content = (
            "%s\tnone\tswap\tsw,comment=cloudconfig\t0\t0\n"
            % (self.swap_path,)
        )

        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab_original_content)

        cc_mounts.handle("", {}, self.mock_cloud, [])

        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    def test_fstab_same_swap_device_already_configured(self):
        """Ensure that cloud-init will not add a swap device if the same
        device already exists in /etc/fstab."""

        fstab_original_content = "%s swap swap defaults 0 0\n" % (
            self.swap_path,
        )
        fstab_expected_content = fstab_original_content

        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab_original_content)

        cc_mounts.handle("", {}, self.mock_cloud, [])

        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    def test_fstab_alternate_swap_device_already_configured(self):
        """Ensure that cloud-init will add a discovered swap device to
        /etc/fstab even when there exists a swap definition on another
        device."""

        fstab_original_content = "/dev/sdc1 swap swap defaults 0 0\n"
        fstab_expected_content = (
            fstab_original_content
            + "%s\tnone\tswap\tsw,comment=cloudconfig\t0\t0\n"
            % (self.swap_path,)
        )

        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab_original_content)

        cc_mounts.handle("", {}, self.mock_cloud, [])

        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_expected_content == fstab_new_content

    def test_no_change_fstab_sets_needs_mount_all(self):
        """verify unchanged fstab entries are mounted if not call mount -a"""
        fstab_original_content = textwrap.dedent(
            f"""
            LABEL=cloudimg-rootfs / ext4 defaults 0 0
            LABEL=UEFI /boot/efi vfat defaults 0 0
            /dev/vdb	/mnt	auto	defaults,noexec,comment=cloudconfig	0	2
            {self.swap_path}	none	swap	sw,comment=cloudconfig	0	0
            """  # noqa: E501
        ).strip()
        cc = {"mounts": [["/dev/vdb", "/mnt", "auto", "defaults,noexec"]]}
        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab_original_content)
        cc_mounts.handle("", cc, self.mock_cloud, [])
        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()
            assert fstab_original_content == fstab_new_content.strip()
        self.m_subp.assert_has_calls(
            [
                mock.call(["mount", "-a"]),
                mock.call(["systemctl", "daemon-reload"]),
            ]
        )

    def test_fstab_mounts_combinations(self):
        """Verify various combinations of mount entries in /etc/fstab."""
        # First and third lines show that even with errors we keep fstab lines
        # unedited unless they contain the cloudconfig comment.
        # 2nd line shows we remove a line with a cloudconfig comment that
        # can be added back in with the mounts config.
        # 4th line shows we remove a line with a cloudconfig comment
        # indiscriminately.
        fstab_original_content = (
            "LABEL=keepme	none	ext4	defaults	0	0\n"
            "/dev/sda1	/a	auto	defaults,comment=cloudconfig	0	2\n"
            "LABEL=UEFI\n"
            "\n"
            "/dev/sda2	/b	auto	defaults,comment=cloudconfig	0	2\n"
        )
        with open(cc_mounts.FSTAB_PATH, "w") as fd:
            fd.write(fstab_original_content)
        cfg = {
            "mounts": [
                # Line that will be overridden due to later None value
                ["/dev/sda3", "dontcare", "auto", "defaults", "0", "0"],
                # Add the one missing default field to the end
                ["/dev/sda4", "/mnt2", "auto", "nofail", "1"],
                # Remove all "/dev/sda3"'s here and earlier
                ["/dev/sda3", None],
                # As long as we have two fields we get the rest of the defaults
                ["/dev/sda5", "/mnt3"],
                # Takes the place of the line that was removed from fstab
                # with the cloudconfig comment
                ["/dev/sda1", "/mnt", "xfs", "auto", None, "2"],
                # The line that survies after previous Nones
                ["/dev/sda3", "/mnt4", "btrfs"],
            ]
        }
        cc_mounts.handle("", cfg, self.mock_cloud, [])
        with open(cc_mounts.FSTAB_PATH, "r") as fd:
            fstab_new_content = fd.read()

        assert (
            fstab_new_content.strip()
            == textwrap.dedent(
                """
            LABEL=keepme	none	ext4	defaults	0	0
            LABEL=UEFI
            /dev/sda4	/mnt2	auto	nofail,comment=cloudconfig	1	2
            /dev/sda5	/mnt3	auto	defaults,nofail,x-systemd.after=cloud-init-network.service,_netdev,comment=cloudconfig	0	2
            /dev/sda1	/mnt	xfs	auto,comment=cloudconfig	0	2
            /dev/sda3	/mnt4	btrfs	defaults,nofail,x-systemd.after=cloud-init-network.service,_netdev,comment=cloudconfig	0	2
            /dev/sdb1	none	swap	sw,comment=cloudconfig	0	0
            """  # noqa: E501
            ).strip()
        )


class TestCreateSwapfile:
    @pytest.mark.parametrize("fstype", ("xfs", "btrfs", "ext4", "other"))
    @mock.patch(M_PATH + "util.get_mount_info")
    @mock.patch(M_PATH + "subp.subp")
    @mock.patch(M_PATH + "util.kernel_version")
    def test_happy_path(
        self, m_kernel_version, m_subp, m_get_mount_info, fstype, tmpdir
    ):
        swap_file = tmpdir.join("swap-file")
        fname = str(swap_file)

        # Some of the calls to subp.subp should create the swap file; this
        # roughly approximates that
        m_subp.side_effect = lambda *args, **kwargs: swap_file.write("")
        m_kernel_version.return_value = (4, 31)

        m_get_mount_info.return_value = (mock.ANY, fstype)

        create_swapfile(fname, "")
        assert mock.call(["mkswap", fname]) in m_subp.call_args_list

    @mock.patch(M_PATH + "util.get_mount_info")
    @mock.patch(M_PATH + "subp.subp")
    def test_fallback_from_fallocate_to_dd(
        self, m_subp, m_get_mount_info, caplog, tmpdir
    ):
        swap_file = tmpdir.join("swap-file")
        fname = str(swap_file)

        def subp_side_effect(cmd, *args, **kwargs):
            # Mock fallocate failing, to initiate fallback
            if cmd[0] == "fallocate":
                raise ProcessExecutionError()

        m_subp.side_effect = subp_side_effect
        # Use ext4 so both fallocate and dd are valid swap creation methods
        m_get_mount_info.return_value = (mock.ANY, "ext4")

        create_swapfile(fname, "")

        cmds = [args[0][0] for args, _kwargs in m_subp.call_args_list]
        assert "fallocate" in cmds, "fallocate was not called"
        assert "dd" in cmds, "fallocate failure did not fallback to dd"

        assert cmds.index("dd") > cmds.index(
            "fallocate"
        ), "dd ran before fallocate"

        assert mock.call(["mkswap", fname]) in m_subp.call_args_list

        msg = "fallocate swap creation failed, will attempt with dd"
        assert msg in caplog.text

    # See https://help.ubuntu.com/community/SwapFaq
    @pytest.mark.parametrize(
        "memsize,expected",
        [
            (256 * MB, 256 * MB),
            (512 * MB, 512 * MB),
            (1 * GB, 1 * GB),
            (2 * GB, 2 * GB),
            (4 * GB, 4 * GB),
            (8 * GB, 4 * GB),
            (16 * GB, 4 * GB),
            (32 * GB, 6 * GB),
            (64 * GB, 8 * GB),
            (128 * GB, 11 * GB),
            (256 * GB, 16 * GB),
            (512 * GB, 23 * GB),
        ],
    )
    def test_suggested_swapsize(self, memsize, expected, mocker):
        mock_stat = namedtuple("mock_stat", "f_frsize f_bfree")
        mocker.patch(
            "os.statvfs",
            # Don't care about available disk space for the purposes of this
            # test
            return_value=mock_stat(math.inf, math.inf),
        )
        size = suggested_swapsize(memsize, math.inf, "dontcare")
        assert expected == approx(size)


class TestMountsSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # We expect to see one mount if provided in user-data.
            (
                {"mounts": []},
                re.escape("mounts: [] ") + test_helpers.SCHEMA_EMPTY_ERROR,
            ),
            # Disallow less than 1 item per mount entry
            (
                {"mounts": [[]]},
                re.escape("mounts.0: [] ") + test_helpers.SCHEMA_EMPTY_ERROR,
            ),
            # Disallow more than 6 items per mount entry
            ({"mounts": [["1"] * 7]}, "mounts.0:.* is too long"),
            # Disallow mount_default_fields will anything other than 6 items
            (
                {"mount_default_fields": ["1"] * 5},
                "mount_default_fields:.* is too short",
            ),
            (
                {"mount_default_fields": ["1"] * 7},
                "mount_default_fields:.* is too long",
            ),
            (
                {"swap": {"invalidprop": True}},
                re.escape(
                    "Additional properties are not allowed ('invalidprop'"
                ),
            ),
            # Swap size/maxsize positive test cases
            ({"swap": {"size": ".5T", "maxsize": ".5T"}}, None),
            ({"swap": {"size": "1G", "maxsize": "1G"}}, None),
            ({"swap": {"size": "200K", "maxsize": "200K"}}, None),
            ({"swap": {"size": "10485760B", "maxsize": "10485760B"}}, None),
            # Swap size/maxsize negative test cases
            ({"swap": {"size": "1.5MB"}}, "swap.size:"),
            (
                {"swap": {"maxsize": "1.5MT"}},
                re.escape("swap.maxsize: '1.5MT' does not match '^([0-9]+)?"),
            ),
            (
                {"swap": {"maxsize": "..5T"}},
                re.escape("swap.maxsize: '..5T' does not match '^([0-9]+)?"),
            ),
            (
                {"swap": {"size": "K"}},
                "swap.size: 'K' is not of type 'integer'",
            ),
        ],
    )
    @test_helpers.skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



import builtins
import glob
import os
import shutil
from pathlib import Path
from unittest import mock

import pytest

from cloudinit import (
    atomic_helper,
    distros,
    helpers,
    lifecycle,
    temp_utils,
)
from cloudinit import user_data as ud
from cloudinit import (
    util,
)
from cloudinit.gpg import GPG
from cloudinit.log import loggers
from tests.unittests.helpers import (
    example_netdev,
    rebase_path,
    retarget_many_wrapper,
)


@pytest.fixture
def Distro(paths):
    def _get_distro(name, cfg=None):
        cls = distros.fetch(name)
        return cls(name, cfg or {}, paths)

    return _get_distro


@pytest.fixture
def m_gpg():
    MockGPG = mock.Mock(spec=GPG)
    MockGPG.configure_mock(**{"getkeybyid.return_value": "fakekey"})
    gpg = MockGPG()
    gpg.list_keys = mock.Mock(return_value="<mocked: list_keys>")
    gpg.getkeybyid = mock.Mock(return_value="<mocked: getkeybyid>")

    # to make tests for cc_apt_configure behave, we need the mocked GPG
    # to actually behave like a context manager
    gpg.__enter__ = GPG.__enter__
    gpg.__exit__ = GPG.__exit__
    yield gpg


FS_FUNCS = {
    os.path: [
        ("isfile", 1),
        ("exists", 1),
        ("islink", 1),
        ("isdir", 1),
        ("lexists", 1),
        ("relpath", 1),
    ],
    os: [
        ("chmod", 2),
        ("chown", 2),
        ("listdir", 1),
        ("lstat", 1),
        ("mkdir", 1),
        ("rename", 2),
        ("rmdir", 1),
        ("scandir", 1),
        ("stat", 1),
        ("symlink", 2),
    ],
    util: [
        ("write_file", 1),
        ("append_file", 1),
        ("load_binary_file", 1),
        ("load_text_file", 1),
        ("ensure_dir", 1),
        ("chmod", 1),
        ("delete_dir_contents", 1),
        ("del_file", 1),
        ("sym_link", -1),
        ("copy", -1),
    ],
    glob: [
        ("glob", 1),
    ],
    builtins: [
        ("open", 1),
    ],
    atomic_helper: [
        ("write_file", 1),
        ("write_json", 1),
    ],
    shutil: [
        ("rmtree", 1),
    ],
}

FS_VARS = {
    temp_utils: ["_ROOT_TMPDIR", "_EXE_ROOT_TMPDIR"],
}


@pytest.fixture
def fake_filesystem_hook():
    """A hook to interact with the real filesystem before mocking it in
    fake_filesystem.

    Fixtures needing to access the real filesystem in tests that use
    fake_filesystem, can depend on this fixture to ensure they run before
    fake_filesystem.

    See in action in tests/unittests/runs/test_simple_run.py.
    """


@pytest.fixture
def fake_filesystem(mocker, tmpdir, fake_filesystem_hook):
    """Mocks fs functions to operate under `tmpdir`

    This fixture is sorted after fix_cloud_init_hook to allow fixtures sorted
    before fake_cloud_init_hook to access the real filesystem.
    """
    # This allows fake_filesystem to be used with production code that
    # creates temporary directories. Functions like TemporaryDirectory()
    # attempt to create a directory under $TMPDIR (among other locations)
    # assuming that it already exists, but then it fails because of the
    # retargeting that happens here.
    TMPDIR = os.getenv("TMPDIR", "/tmp")
    Path(tmpdir, TMPDIR[1:]).mkdir(parents=True, exist_ok=True)
    Path(tmpdir, "tmp").mkdir(exist_ok=True)

    for mod, funcs in FS_FUNCS.items():
        for f, nargs in funcs:
            func = getattr(mod, f)
            trap_func = retarget_many_wrapper(str(tmpdir), nargs, func)
            mocker.patch.object(mod, f, trap_func)

    for mod, vars in FS_VARS.items():
        for var_name in vars:
            var_val = getattr(mod, var_name)
            new_var_val = rebase_path(var_val, str(tmpdir))
            mocker.patch.object(mod, var_name, new_var_val)

    yield str(tmpdir)


@pytest.fixture(scope="session", autouse=True)
def disable_sysfs_net(tmpdir_factory):
    """Avoid tests which read the underlying host's /syc/class/net."""
    mock_sysfs = f"{tmpdir_factory.mktemp('sysfs')}/"
    with mock.patch(
        "cloudinit.net.get_sys_class_path", return_value=mock_sysfs
    ):
        yield mock_sysfs


@pytest.fixture(scope="class")
def disable_netdev_info(request):
    """Avoid tests which read the underlying host's /syc/class/net."""
    with mock.patch(
        "cloudinit.netinfo.netdev_info", return_value=example_netdev
    ) as mock_netdev:
        yield mock_netdev


@pytest.fixture(autouse=True)
def disable_dns_lookup(request):
    if "allow_dns_lookup" in request.keywords:
        yield
        return

    def side_effect(args, *other_args, **kwargs):
        raise AssertionError("Unexpectedly used util.is_resolvable")

    with mock.patch(
        "cloudinit.util.is_resolvable", side_effect=side_effect, autospec=True
    ):
        yield


@pytest.fixture()
def dhclient_exists():
    with mock.patch(
        "cloudinit.net.dhcp.subp.which",
        return_value="/sbin/dhclient",
        autospec=True,
    ):
        yield


loggers.configure_root_logger()


@pytest.fixture(autouse=True, scope="session")
def disable_root_logger_setup():
    with mock.patch(
        "cloudinit.log.loggers.configure_root_logger", autospec=True
    ):
        yield


@pytest.fixture
def clear_deprecation_log():
    """Clear any deprecation warnings before and after running tests."""
    # Since deprecations are de-duped, the existence (or non-existence) of
    # a deprecation warning in a previous test can cause the next test to
    # fail.
    setattr(lifecycle.deprecate, "log", set())


PYTEST_VERSION_TUPLE = tuple(map(int, pytest.__version__.split(".")))

if PYTEST_VERSION_TUPLE < (3, 9, 0):

    @pytest.fixture
    def tmp_path(tmpdir):
        return Path(tmpdir)


@pytest.fixture
def paths(tmpdir) -> helpers.Paths:
    """
    Return a helpers.Paths object configured to use a tmpdir.

    (This uses the builtin tmpdir fixture.)
    """
    dirs = {
        "cloud_dir": tmpdir.mkdir("cloud_dir").strpath,
        "docs_dir": tmpdir.mkdir("docs_dir").strpath,
        "run_dir": tmpdir.mkdir("run_dir").strpath,
        "templates_dir": tmpdir.mkdir("templates_dir").strpath,
    }
    return helpers.Paths(dirs)


@pytest.fixture
def ud_proc(paths):
    return ud.UserDataProcessor(paths)


@pytest.fixture
def socket_attrs(mocker):
    """A fixture to add to some attributes to the socket module.

    Many socket attributes are OS-specific, so ensure we have attributes
    that work for the tests that need them.
    """
    mocker.patch("socket.AF_NETLINK", 0, create=True)
    mocker.patch("socket.NETLINK_ROUTE", 0, create=True)
    mocker.patch("socket.SOCK_CLOEXEC", 0, create=True)


@pytest.fixture
def fake_socket(mocker, socket_attrs):
    """A fixture to mock socket.socket()."""
    # Even though this is just a one-liner, if we need to mock
    # socket.socket, we want to ensure that socket_attrs is
    # applied too.
    return mocker.patch("socket.socket", autospec=True)



from pathlib import Path

import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.util import (
    ASSETS_DIR,
    verify_clean_boot,
    verify_clean_log,
)

PICKLE_PATH = Path("/var/lib/cloud/instance/obj.pkl")
TEST_PICKLE = ASSETS_DIR / "test_version_change.pkl"


def _assert_no_pickle_problems(log, client):
    assert "Failed loading pickled blob" not in log
    verify_clean_log(log)
    verify_clean_boot(client)


def test_reboot_without_version_change(client: IntegrationInstance):
    log = client.read_from_file("/var/log/cloud-init.log")
    assert "Python version change detected" not in log
    assert "Cache compatibility status is currently unknown." not in log
    _assert_no_pickle_problems(log, client)

    client.restart()
    log = client.read_from_file("/var/log/cloud-init.log")
    assert "Python version change detected" not in log
    assert "Could not determine Python version used to write cache" not in log
    _assert_no_pickle_problems(log, client)

    # Now ensure that loading a bad pickle gives us problems
    client.push_file(TEST_PICKLE, PICKLE_PATH)
    client.restart()
    log = client.read_from_file("/var/log/cloud-init.log")

    # no cache found is an "expected" upgrade error, and
    # "Failed" means we're unable to load the pickle
    assert any(
        [
            "Failed loading pickled blob from {}".format(PICKLE_PATH) in log,
            "no cache found" in log,
        ]
    )


@pytest.mark.skipif(
    PLATFORM
    not in ["ec2", "gce", "oci", "openstack", "lxd_container", "lxd_vm"],
    reason=f"Test hasn't been tested on {PLATFORM}.",
)
# TODO: The below comment likely isn't true anymore
# No Azure because the cache gets purged every reboot, so we'll never
# get to the point where we need to purge cache due to version change
def test_cache_purged_on_version_change(client: IntegrationInstance):
    # Start by pushing the invalid pickle so we'll hit an error if the
    # cache didn't actually get purged
    client.push_file(TEST_PICKLE, PICKLE_PATH)
    client.execute("echo '1.0' > /var/lib/cloud/data/python-version")
    client.restart()
    log = client.read_from_file("/var/log/cloud-init.log")
    assert "Python version change detected. Purging cache" in log
    _assert_no_pickle_problems(log, client)


def test_log_message_on_missing_version_file(client: IntegrationInstance):
    client.execute("rm /var/lib/cloud/data/python-version")
    client.execute("rm /var/log/cloud-init.log")
    client.restart()
    log = client.read_from_file("/var/log/cloud-init.log")
    assert (
        "Writing python-version file. "
        "Cache compatibility status is currently unknown." in log
    )



"""Defines _Reaper, which destroys instances in a background thread

This class is intended to be a singleton which is instantiated on session setup
and cleaned on session teardown. Any instances submitted to the reaper are
destroyed. Instances that refuse to be destroyed due to external library errors
or flaky infrastructure are tracked, retried and upon test session completion
are reported to the end user as a test warning.
"""

import logging
import queue
import threading
import warnings
from typing import Final, List, Optional

from tests.integration_tests.instances import IntegrationInstance

LOG = logging.getLogger()


class Reaper:
    def __init__(self, timeout: float = 30.0):
        # self.timeout sets the amount of time to sleep before retrying
        self.timeout = timeout
        # self.wake_reaper tells the reaper to wake up.
        #
        # A lock is used for synchronization. This means that notify() will
        # block if
        # the reaper is currently awake.
        #
        # It is set by:
        # - signal interrupt indicating cleanup
        # - session completion indicating cleanup
        # - reaped instance indicating work to be done
        self.wake_reaper: Final[threading.Condition] = threading.Condition()

        # self.exit_reaper tells the reaper loop to tear down, called once at
        # end of tests
        self.exit_reaper: Final[threading.Event] = threading.Event()

        # List of instances which temporarily escaped death
        # The primary purpose of the reaper is to coax these instance towards
        # eventual demise and report their insubordination on shutdown.
        self.undead_ledger: Final[List[IntegrationInstance]] = []

        # Queue of newly reaped instances
        self.reaped_instances: Final[queue.Queue[IntegrationInstance]] = (
            queue.Queue()
        )

        # Thread object, handle used to re-join the thread
        self.reaper_thread: Optional[threading.Thread] = None

        # Count the dead
        self.counter = 0

    def reap(self, instance: IntegrationInstance):
        """reap() submits an instance to the reaper thread.

        An instance that is passed to the reaper must not be used again. It may
        not be dead yet, but it has no place among the living.
        """
        LOG.info("Reaper: receiving %s", instance.instance.id)

        self.reaped_instances.put(instance)
        with self.wake_reaper:
            self.wake_reaper.notify()
            LOG.info("Reaper: awakened to reap")

    def start(self):
        """Spawn the reaper background thread."""
        LOG.info("Reaper: starting")
        self.reaper_thread = threading.Thread(
            target=self._reaper_loop, name="reaper"
        )
        self.reaper_thread.start()

    def stop(self):
        """Stop the reaper background thread and wait for completion."""
        LOG.info("Reaper: stopping")
        self.exit_reaper.set()
        with self.wake_reaper:
            self.wake_reaper.notify()
            LOG.info("Reaper: awakened to reap")
        if self.reaper_thread and self.reaper_thread.is_alive():
            self.reaper_thread.join()
        LOG.info("Reaper: stopped")

    def _destroy(self, instance: IntegrationInstance) -> bool:
        """destroy() destroys an instance and returns True on success."""
        try:
            LOG.info("Reaper: destroying %s", instance.instance.id)
            instance.destroy()
            self.counter += 1
            return True
        except Exception as e:
            LOG.warning(
                "Error while tearing down instance %s: %s ", instance, e
            )
            return False

    def _reaper_loop(self) -> None:
        """reaper_loop() manages all instances that have been reaped

        tasks:
        - destroy newly reaped instances
        - manage a ledger undead instances
        - periodically attempt to kill undead instances
        - die when instructed to
        - ensure that every reaped instance is destroyed at least once before
          reaper dies
        """
        LOG.info("Reaper: exalted in life, to assist others in death")
        while True:
            # nap until woken or timeout
            with self.wake_reaper:
                self.wake_reaper.wait(timeout=self.timeout)
            if self._do_reap():
                break
        LOG.info("Reaper: exited")

    def _do_reap(self) -> bool:
        """_do_reap does a single pass of the reaper loop

        return True if the loop should exit
        """

        new_undead_instances: List[IntegrationInstance] = []

        # first destroy all newly reaped instances
        while not self.reaped_instances.empty():
            instance = self.reaped_instances.get_nowait()
            instance_id = instance.instance.id
            success = self._destroy(instance)
            if not success:
                LOG.warning(
                    "Reaper: failed to destroy %s",
                    instance.instance.id,
                )
                # failure to delete, add to the ledger
                new_undead_instances.append(instance)
            else:
                LOG.info("Reaper: destroyed %s", instance_id)

        # every instance has tried at least once and the reaper has been
        # instructed to tear down - so do it
        if self.exit_reaper.is_set():
            if not self.reaped_instances.empty():
                # race: an instance was added to the queue after iteration
                # completed. Destroy the latest instance.
                self._update_undead_ledger(new_undead_instances)
                return False
            self._update_undead_ledger(new_undead_instances)
            LOG.info("Reaper: exiting")
            if self.undead_ledger:
                # undead instances exist - unclean teardown
                LOG.info(
                    "Reaper: the faults of incompetent abilities will be "
                    "consigned to oblivion, as myself must soon be to the "
                    "mansions of rest."
                )
                warnings.warn(f"Test instance(s) leaked: {self.undead_ledger}")
            else:
                LOG.info("Reaper: duties complete, my turn to rest")
            LOG.info(
                "Reaper: reaped %s/%s instances",
                self.counter,
                self.counter + len(self.undead_ledger),
            )
            return True

        # attempt to destroy all instances which previously refused to destroy
        for instance in self.undead_ledger:
            if self.exit_reaper.is_set() and self.reaped_instances.empty():
                # don't retry instances if the exit_reaper Event is set
                break
            instance_id = instance.instance.id
            if self._destroy(instance):
                self.undead_ledger.remove(instance)
                LOG.info("Reaper: destroyed %s (undead)", instance_id)

        self._update_undead_ledger(new_undead_instances)
        return False

    def _update_undead_ledger(
        self, new_undead_instances: List[IntegrationInstance]
    ):
        """update the ledger with newly undead instances"""
        if new_undead_instances:
            if self.undead_ledger:
                LOG.info(
                    "Reaper: instance(s) not ready to die %s, will now join "
                    "the ranks of the undead: %s",
                    new_undead_instances,
                    self.undead_ledger,
                )
            else:
                LOG.info(
                    "Reaper: instance(s) not ready to die %s",
                    new_undead_instances,
                )
        self.undead_ledger.extend(new_undead_instances)
        return False



# This file is part of cloud-init. See LICENSE file for license information.

from tests.unittests.helpers import get_distro, mock

M_PATH = "cloudinit.distros.openbsd."


class TestOpenBSD:
    @mock.patch(M_PATH + "subp.subp")
    def test_add_user(self, m_subp):
        distro = get_distro("openbsd")
        assert True is distro.add_user("me2", uid=1234, default=False)
        assert [
            mock.call(
                ["useradd", "-m", "me2"], logstring=["useradd", "-m", "me2"]
            )
        ] == m_subp.call_args_list

    def test_unlock_passwd(self, caplog):
        distro = get_distro("openbsd")
        distro.unlock_passwd("me2")
        assert (
            "OpenBSD password lock is not reversible, "
            "ignoring unlock for user me2" in caplog.text
        )



# This file is part of cloud-init. See LICENSE file for license information.
import base64
import json
import os
import re
import struct
import time
import uuid
import zlib
from unittest import mock

import pytest

from cloudinit.reporting import events, instantiated_handler_registry
from cloudinit.reporting.handlers import HyperVKvpReportingHandler
from cloudinit.sources.helpers import azure


class TestKvpEncoding:
    def test_encode_decode(self):
        kvp = {"key": "key1", "value": "value1"}
        kvp_reporting = HyperVKvpReportingHandler()
        data = kvp_reporting._encode_kvp_item(kvp["key"], kvp["value"])
        assert len(data) == kvp_reporting.HV_KVP_RECORD_SIZE
        decoded_kvp = kvp_reporting._decode_kvp_item(data)
        assert kvp == decoded_kvp


class TestKvpReporter:
    @pytest.fixture(autouse=True)
    def mock_dmi(self, mocker):
        """Mock dmi.read_dmi_data to prevent subp calls from vm_id property."""
        mocker.patch(
            "cloudinit.reporting.handlers.dmi.read_dmi_data",
            return_value=None,
        )

    @pytest.fixture
    def kvp_file_path(self, tmp_path):
        file_path = tmp_path / "kvp_pool_file"
        file_path.touch()
        return str(file_path)

    @pytest.fixture
    def reporter(self, kvp_file_path):
        return HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)

    def test_events_with_higher_incarnation_not_over_written(
        self, kvp_file_path, reporter
    ):
        assert 0 == len(list(reporter._iterate_kvps(0)))

        reporter.publish_event(
            events.ReportingEvent("foo", "name1", "description")
        )
        reporter.publish_event(
            events.ReportingEvent("foo", "name2", "description")
        )
        reporter.q.join()
        assert 2 == len(list(reporter._iterate_kvps(0)))

        reporter3 = HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)
        reporter3.incarnation_no = reporter.incarnation_no - 1
        reporter3.publish_event(
            events.ReportingEvent("foo", "name3", "description")
        )
        reporter3.q.join()
        assert 3 == len(list(reporter3._iterate_kvps(0)))

    def test_finish_event_result_is_logged(self, reporter):
        reporter.publish_event(
            events.FinishReportingEvent(
                "name2",
                "description1",
                duration=2.5,
                result=events.status.FAIL,
            )
        )
        reporter.q.join()
        kvp_value = json.loads(list(reporter._iterate_kvps(0))[0]["value"])
        assert kvp_value["result"] == events.status.FAIL
        assert kvp_value["duration"] == 2.5

    def test_file_operation_issue(self, kvp_file_path, reporter):
        os.remove(kvp_file_path)
        reporter.publish_event(
            events.FinishReportingEvent(
                "name2",
                "description1",
                duration=1.0,
                result=events.status.FAIL,
            )
        )
        reporter.q.join()

    def test_event_very_long(self, reporter):
        description = "ab" * reporter.HV_KVP_AZURE_MAX_VALUE_SIZE
        long_event = events.FinishReportingEvent(
            "event_name",
            description,
            duration=3.0,
            result=events.status.FAIL,
        )
        reporter.publish_event(long_event)
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 3 == len(kvps)

        # restore from the kvp to see the content are all there
        full_description = ""
        for i in range(len(kvps)):
            msg_slice = json.loads(kvps[i]["value"])
            assert msg_slice["msg_i"] == i
            assert msg_slice["duration"] == 3.0
            full_description += msg_slice["msg"]
        assert description == full_description

    def test_not_truncate_kvp_file_modified_after_boot(
        self, kvp_file_path, reporter
    ):
        with open(kvp_file_path, "wb+") as f:
            kvp = {"key": "key1", "value": "value1"}
            data = struct.pack(
                "%ds%ds"
                % (
                    HyperVKvpReportingHandler.HV_KVP_EXCHANGE_MAX_KEY_SIZE,
                    HyperVKvpReportingHandler.HV_KVP_EXCHANGE_MAX_VALUE_SIZE,
                ),
                kvp["key"].encode("utf-8"),
                kvp["value"].encode("utf-8"),
            )
            f.write(data)
        cur_time = time.time()
        os.utime(kvp_file_path, (cur_time, cur_time))

        # reset this because the unit test framework
        # has already polluted the class variable
        HyperVKvpReportingHandler._already_truncated_pool_file = False

        kvps = list(reporter._iterate_kvps(0))
        assert 1 == len(kvps)

    def test_truncate_stale_kvp_file(self, kvp_file_path, reporter):
        with open(kvp_file_path, "wb+") as f:
            kvp = {"key": "key1", "value": "value1"}
            data = struct.pack(
                "%ds%ds"
                % (
                    HyperVKvpReportingHandler.HV_KVP_EXCHANGE_MAX_KEY_SIZE,
                    HyperVKvpReportingHandler.HV_KVP_EXCHANGE_MAX_VALUE_SIZE,
                ),
                kvp["key"].encode("utf-8"),
                kvp["value"].encode("utf-8"),
            )
            f.write(data)

        # set the time ways back to make it look like
        # we had an old kvp file
        os.utime(kvp_file_path, (1000000, 1000000))

        # reset this because the unit test framework
        # has already polluted the class variable
        HyperVKvpReportingHandler._already_truncated_pool_file = False

        reporter = HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)
        kvps = list(reporter._iterate_kvps(0))
        assert 0 == len(kvps)

    @mock.patch("cloudinit.distros.uses_systemd")
    @mock.patch("cloudinit.subp.subp")
    def test_get_boot_telemetry(self, m_subp, m_sysd, reporter):
        datetime_pattern = (
            r"\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]"
            r"\d:[0-5]\d\.\d+([+-][0-2]\d:[0-5]\d|Z)"
        )

        # get_boot_telemetry makes two subp calls to systemctl. We provide
        # a list of values that the subp calls should return
        m_subp.side_effect = [
            ("UserspaceTimestampMonotonic=1844838", ""),
            ("InactiveExitTimestampMonotonic=3068203", ""),
        ]
        m_sysd.return_value = True

        reporter.publish_event(azure.get_boot_telemetry())
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 1 == len(kvps)

        evt_msg = kvps[0]["value"]
        if not re.search("kernel_start=" + datetime_pattern, evt_msg):
            raise AssertionError("missing kernel_start timestamp")
        if not re.search("user_start=" + datetime_pattern, evt_msg):
            raise AssertionError("missing user_start timestamp")
        if not re.search("cloudinit_activation=" + datetime_pattern, evt_msg):
            raise AssertionError("missing cloudinit_activation timestamp")

    def test_get_system_info(self, reporter):
        pattern = r"[^=\s]+"

        reporter.publish_event(azure.get_system_info())
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 1 == len(kvps)
        evt_msg = kvps[0]["value"]

        # the most important information is cloudinit version,
        # kernel_version, and the distro variant. It is ok if
        # if the rest is not available
        if not re.search("cloudinit_version=" + pattern, evt_msg):
            raise AssertionError("missing cloudinit_version string")
        if not re.search("kernel_version=" + pattern, evt_msg):
            raise AssertionError("missing kernel_version string")
        if not re.search("variant=" + pattern, evt_msg):
            raise AssertionError("missing distro variant string")

    def test_report_diagnostic_event_without_logger_func(self, reporter):
        diagnostic_msg = "test_diagnostic"
        reporter.publish_event(azure.report_diagnostic_event(diagnostic_msg))
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 1 == len(kvps)
        evt_msg = kvps[0]["value"]

        if diagnostic_msg not in evt_msg:
            raise AssertionError("missing expected diagnostic message")

    def test_report_diagnostic_event_with_logger_func(self, reporter):
        logger_func = mock.MagicMock()
        diagnostic_msg = "test_diagnostic"
        reporter.publish_event(
            azure.report_diagnostic_event(
                diagnostic_msg, logger_func=logger_func
            )
        )
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 1 == len(kvps)
        evt_msg = kvps[0]["value"]

        if diagnostic_msg not in evt_msg:
            raise AssertionError("missing expected diagnostic message")
        logger_func.assert_called_once_with(diagnostic_msg)

    def test_report_compressed_event(self, reporter):
        try:
            instantiated_handler_registry.register_item("telemetry", reporter)
            event_desc = b"test_compressed"
            azure.report_compressed_event("compressed event", event_desc)

            self.validate_compressed_kvps(reporter, 1, [event_desc])
        finally:
            instantiated_handler_registry.unregister_item(
                "telemetry", force=False
            )

    def validate_compressed_kvps(self, reporter, count, values):
        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        compressed_count = 0
        for i in range(len(kvps)):
            kvp = kvps[i]
            kvp_value = kvp["value"]
            kvp_value_json = json.loads(kvp_value)
            evt_msg = kvp_value_json["msg"]
            evt_type = kvp_value_json["type"]
            if evt_type != azure.COMPRESSED_EVENT_TYPE:
                continue
            evt_msg_json = json.loads(evt_msg)
            evt_encoding = evt_msg_json["encoding"]
            evt_data = zlib.decompress(
                base64.decodebytes(evt_msg_json["data"].encode("ascii"))
            )

            assert compressed_count < len(values)
            assert evt_data == values[compressed_count]
            assert evt_encoding == "gz+b64"
            compressed_count += 1
        assert compressed_count == count

    def test_unique_kvp_key(self, reporter):
        evt1 = events.ReportingEvent(
            "event_type", "event_message", "event_description"
        )
        reporter.publish_event(evt1)

        evt2 = events.ReportingEvent(
            "event_type",
            "event_message",
            "event_description",
            timestamp=evt1.timestamp + 1,
        )
        reporter.publish_event(evt2)

        reporter.q.join()
        kvps = list(reporter._iterate_kvps(0))
        assert 2 == len(kvps)
        assert (
            kvps[0]["key"] != kvps[1]["key"]
        ), "duplicate keys for KVP entries"

    def test_write_key(self, reporter):
        reporter.write_key("test-key", "test-value")
        assert list(reporter._iterate_kvps(0)) == [
            {"key": "test-key", "value": "test-value"}
        ]

    def test_write_key_truncates(self, reporter):
        value = "A" * 2000
        reporter.write_key("test-key", value)

        assert len(list(reporter._iterate_kvps(0))[0]["value"]) == 1023

    def test_vm_id_defaults_to_zero_guid_when_dmi_fails(self, kvp_file_path):
        """Test handler defaults to ZERO_GUID when DMI query fails."""
        with mock.patch(
            "cloudinit.reporting.handlers.dmi.read_dmi_data",
            side_effect=RuntimeError("not available"),
        ):
            reporter = HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)
            assert reporter.vm_id == HyperVKvpReportingHandler.ZERO_GUID

    def test_vm_id_fallback_reads_system_uuid(self, kvp_file_path):
        """Handler falls back to DMI system-uuid when vm_id not set."""
        with mock.patch(
            "cloudinit.reporting.handlers.dmi.read_dmi_data",
            return_value="AABBCCDD-EEFF-0011-2233-445566778899",
        ) as m_read_dmi:
            reporter = HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)
            assert reporter.vm_id == "aabbccdd-eeff-0011-2233-445566778899"
            m_read_dmi.assert_called_once_with("system-uuid")

    def test_vm_id_keeps_zero_guid_when_system_uuid_missing(
        self, kvp_file_path
    ):
        """If DMI system-uuid isn't available, vm_id remains ZERO_GUID."""
        with mock.patch(
            "cloudinit.reporting.handlers.dmi.read_dmi_data", return_value=None
        ) as m_read_dmi:
            reporter = HyperVKvpReportingHandler(kvp_file_path=kvp_file_path)
            assert reporter.vm_id == HyperVKvpReportingHandler.ZERO_GUID
            m_read_dmi.assert_called_once_with("system-uuid")

    def test_event_key_format(self, reporter):
        """Event key must end with vm_id followed by a uuid."""
        evt = events.ReportingEvent("type", "name", "description")
        reporter.publish_event(evt)
        reporter.q.join()

        key_parts = list(reporter._iterate_kvps(0))[0]["key"].split("|")
        assert len(key_parts) == 6

        assert key_parts[0] == HyperVKvpReportingHandler.EVENT_PREFIX
        assert key_parts[1] == str(reporter.incarnation_no)
        assert key_parts[2] == evt.event_type
        assert key_parts[3] == evt.name
        assert key_parts[4] == reporter.vm_id
        uuid_part = key_parts[5]
        assert uuid.UUID(uuid_part)
        assert uuid_part != reporter.vm_id



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Cosmin Luta
# Copyright (C) 2012 Yahoo! Inc.
# Copyright (C) 2012 Gerard Dethier
# Copyright (C) 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Cosmin Luta <q4break@gmail.com>
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
# Author: Gerard Dethier <g.dethier@gmail.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
import time
from contextlib import suppress
from socket import gaierror, getaddrinfo, inet_ntoa
from struct import pack

from cloudinit import dmi, net, performance, sources
from cloudinit import url_helper as uhelp
from cloudinit import util
from cloudinit.net import dhcp
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.net.ephemeral import EphemeralIPNetwork
from cloudinit.sources.helpers import ec2

LOG = logging.getLogger(__name__)

CLOUD_STACK_DMI_NAME = "CloudStack"


class CloudStackPasswordServerClient:
    """
    Implements password fetching from the CloudStack password server.

    http://cloudstack-administration.readthedocs.org/
       en/latest/templates.html#adding-password-management-to-your-templates
    has documentation about the system.  This implementation is following that
    found at
    https://github.com/shankerbalan/cloudstack-scripts/
       blob/master/cloud-set-guest-password-debian
    """

    def __init__(self, virtual_router_address):
        self.virtual_router_address = virtual_router_address

    def _do_request(self, domu_request):
        url = f"http://{self.virtual_router_address}:8080"
        headers = {"DomU_Request": domu_request}

        resp = uhelp.readurl(url, headers=headers, timeout=20, retries=3)

        if not resp.ok():
            raise RuntimeError("Failed to fetch VM password from CloudStack")

        return resp.contents.decode("utf-8").strip()

    @performance.timed("Getting password", log_mode="always")
    def get_password(self):
        password = self._do_request("send_my_password")
        if password in ["", "saved_password"]:
            return None
        if password == "bad_request":
            raise RuntimeError("Error when attempting to fetch root password.")
        self._do_request("saved_password")
        return password


class DataSourceCloudStack(sources.DataSource):
    perform_dhcp_setup = False
    dsname = "CloudStack"

    # Setup read_url parameters per get_url_params.
    url_max_wait = 120
    url_timeout = 50

    def __init__(self, sys_cfg, distro, paths):
        super().__init__(sys_cfg, distro, paths)
        self.seed_dir = os.path.join(paths.seed_dir, "cs")
        # Cloudstack has its metadata/userdata URLs located at
        # http://<virtual-router-ip>/latest/
        self.api_ver = "latest"
        self.cfg = {}
        self.vr_addr = None

    def _get_domainname(self):
        """Try obtaining a "domain-name" DHCP lease parameter:
        - From systemd-networkd lease (case-insensitive)
        - From ISC dhclient
        - From dhcpcd (ephemeral)
        - Return empty string if not found (non-fatal)
        """

        LOG.debug("Try obtaining domain name from networkd leases")
        for key in ["DOMAINNAME", "Domain", "domain-name"]:
            domainname = dhcp.networkd_get_option_from_leases(key)
            if domainname:
                return domainname.strip()

        LOG.debug(
            "Could not obtain FQDN from networkd leases. Falling back to "
            "ISC dhclient"
        )
        with suppress(dhcp.NoDHCPLeaseMissingDhclientError):
            domain_name = dhcp.IscDhclient().get_key_from_latest_lease(
                self.distro, "domain-name"
            )
            if domain_name:
                return domain_name.strip()

        LOG.debug(
            "Could not obtain FQDN from ISC dhclient leases. Falling back to "
            "%s",
            self.distro.dhcp_client.client_name,
        )
        try:
            latest_lease = self.distro.dhcp_client.get_newest_lease(
                self.distro.fallback_interface
            )
            domain_name = latest_lease.get("domain-name")
            if domain_name:
                return domain_name.strip()
        except (NoDHCPLeaseError, FileNotFoundError, AttributeError):
            pass

        LOG.debug("No domain name found in any DHCP lease; returning empty")
        return ""

    def get_hostname(
        self,
        fqdn=False,
        resolve_ip=False,
        metadata_only=False,
    ):
        """
        Returns instance's hostname / fqdn
        First probes the parent class method.

        If fqdn is requested, and the parent method didn't return it,
        then attach the domain-name from DHCP response.
        """
        hostname = super().get_hostname(fqdn, resolve_ip, metadata_only)
        if fqdn and "." not in hostname.hostname:
            LOG.debug("FQDN requested")
            domainname = self._get_domainname()
            if domainname:
                fqdn = f"{hostname.hostname}.{domainname}"
                LOG.debug("Obtained the following FQDN: %s", fqdn)
                return sources.DataSourceHostname(fqdn, hostname.is_default)
            LOG.debug(
                "Could not determine domain name for FQDN. "
                "Fall back to hostname as an FQDN: %s",
                fqdn,
            )
        return hostname

    def wait_for_metadata_service(self):
        url_params = self.get_url_params()

        if url_params.max_wait_seconds <= 0:
            return False

        urls = [
            uhelp.combine_url(
                self.metadata_address, "latest/meta-data/instance-id"
            )
        ]
        start_time = time.monotonic()
        url, _response = uhelp.wait_for_url(
            urls=urls,
            max_wait=url_params.max_wait_seconds,
            timeout=url_params.timeout_seconds,
            status_cb=LOG.warning,
        )

        if url:
            LOG.debug("Using metadata source: '%s'", url)
        else:
            LOG.critical(
                "Giving up on waiting for the metadata from %s"
                " after %s seconds",
                urls,
                int(time.monotonic() - start_time),
            )

        return bool(url)

    def get_config_obj(self):
        return self.cfg

    @staticmethod
    def ds_detect() -> bool:
        """Check if running on this datasource"""
        return is_platform_viable()

    def _get_data(self):
        seed_ret = {}
        if util.read_optional_seed(seed_ret, base=(self.seed_dir + "/")):
            self.userdata_raw = seed_ret["user-data"]
            self.metadata = seed_ret["meta-data"]
            LOG.debug("Using seeded cloudstack data from: %s", self.seed_dir)
            return True
        if self.perform_dhcp_setup:
            primary_nic = net.find_fallback_nic()
            LOG.debug("Attempting DHCP on: %s", primary_nic)
            network_context = EphemeralIPNetwork(self.distro, primary_nic)
        else:
            network_context = util.nullcontext()
        try:
            with network_context:
                vr_addr = get_vr_address(self.distro)
                # If vr_addr is a dict, we have the DHCP lease
                self.vr_addr = (
                    vr_addr.get("dhcp-server-identifier")
                    if isinstance(vr_addr, dict)
                    else vr_addr
                )
                if not self.vr_addr:
                    raise RuntimeError("No virtual router found!")
                self.metadata_address = f"http://{self.vr_addr}/"
                if not self.wait_for_metadata_service():
                    return False

                return self._crawl_metadata()
        except NoDHCPLeaseError:
            LOG.warning("Unable to obtain a DHCP lease on %s", primary_nic)
            return False
        except Exception as e:
            LOG.warning("Failed fetching metadata service: %s", str(e))
            return False

    @performance.timed("Crawling metadata", log_mode="always")
    def _crawl_metadata(self):
        self.userdata_raw = ec2.get_instance_userdata(
            self.api_ver, self.metadata_address
        )
        self.metadata = ec2.get_instance_metadata(
            self.api_ver, self.metadata_address
        )

        password_client = CloudStackPasswordServerClient(self.vr_addr)
        try:
            set_password = password_client.get_password()
        except Exception:
            util.logexc(
                LOG,
                "Failed to fetch password from virtual router %s",
                self.vr_addr,
            )
        else:
            if set_password:
                self.cfg = {
                    "ssh_pwauth": True,
                    "password": set_password,
                    "chpasswd": {
                        "expire": False,
                    },
                }
        return True

    def get_instance_id(self):
        return self.metadata["instance-id"]

    @property
    def availability_zone(self):
        return self.metadata["availability-zone"]


class DataSourceCloudStackLocal(DataSourceCloudStack):
    """Run in init-local using a dhcp discovery prior to metadata crawl.

    In init-local, no network is available. This subclass sets up minimal
    networking with dhclient on a viable nic so that it can talk to the
    metadata service. If the metadata service provides network configuration
    then render the network configuration for that instance based on metadata.
    """

    perform_dhcp_setup = True  # Get metadata network config if present


def get_data_server():
    # Returns the metadataserver from dns
    try:
        addrinfo = getaddrinfo("data-server", 80)
    except gaierror:
        LOG.debug("DNS Entry data-server not found")
        return None
    else:
        return addrinfo[0][4][0]  # return IP


def get_default_gateway():
    # Returns the default gateway ip address in the dotted format.
    lines = util.load_text_file("/proc/net/route").splitlines()
    for line in lines:
        items = line.split("\t")
        if items[1] == "00000000":
            # Found the default route, get the gateway
            gw = inet_ntoa(pack("<L", int(items[2], 16)))
            LOG.debug("Found default route, gateway is %s", gw)
            return gw
    return None


def get_vr_address(distro):
    # Get the address of the virtual router via dhcp leases
    # If no virtual router is detected, fallback on default gateway.
    # See http://docs.cloudstack.apache.org/projects/cloudstack-administration/en/4.8/virtual_machines/user-data.html # noqa

    # Try data-server DNS entry first
    latest_address = get_data_server()
    if latest_address:
        LOG.debug(
            "Found metadata server '%s' via data-server DNS entry",
            latest_address,
        )
        return latest_address

    # Try networkd second...
    latest_address = dhcp.networkd_get_option_from_leases("SERVER_ADDRESS")
    if latest_address:
        LOG.debug(
            "Found SERVER_ADDRESS '%s' via networkd_leases", latest_address
        )
        return latest_address

    # Try dhcp lease files next
    # get_key_from_latest_lease() needs a Distro object to know which directory
    # stores lease files
    with suppress(dhcp.NoDHCPLeaseMissingDhclientError):
        latest_address = dhcp.IscDhclient().get_key_from_latest_lease(
            distro, "dhcp-server-identifier"
        )
        if latest_address:
            LOG.debug("Found SERVER_ADDRESS '%s' via dhclient", latest_address)
            return latest_address

    with suppress(FileNotFoundError):
        latest_lease = distro.dhcp_client.get_newest_lease(
            distro.fallback_interface
        )
        if latest_lease:
            LOG.debug(
                "Found SERVER_ADDRESS '%s' via ephemeral %s lease ",
                latest_lease,
                distro.dhcp_client.client_name,
            )
            return latest_lease

    # No virtual router found, fallback to default gateway
    LOG.debug("No DHCP found, using default gateway")
    return get_default_gateway()


def is_platform_viable() -> bool:
    product_name = dmi.read_dmi_data("system-product-name")
    if not product_name:
        LOG.debug("system-product-name not available in dmi data")
        return False
    return product_name.startswith(CLOUD_STACK_DMI_NAME)


# Used to match classes to dependencies
datasources = [
    (DataSourceCloudStackLocal, (sources.DEP_FILESYSTEM,)),
    (DataSourceCloudStack, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

import pytest

from cloudinit.registry import DictRegistry


class TestDictRegistry:
    def test_added_item_included_in_output(self):
        registry = DictRegistry()
        item_key, item_to_register = "test_key", mock.Mock()
        registry.register_item(item_key, item_to_register)
        assert {item_key: item_to_register} == registry.registered_items

    def test_registry_starts_out_empty(self):
        assert {} == DictRegistry().registered_items

    def test_modifying_registered_items_isnt_exposed_to_other_callers(self):
        registry = DictRegistry()
        registry.registered_items["test_item"] = mock.Mock()
        assert {} == registry.registered_items

    def test_keys_cannot_be_replaced(self):
        registry = DictRegistry()
        item_key = "test_key"
        registry.register_item(item_key, mock.Mock())

        with pytest.raises(ValueError):
            registry.register_item(item_key, mock.Mock())



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

from cloudinit import version


class TestExportsFeatures:
    def test_has_network_config_v1(self):
        assert "NETWORK_CONFIG_V1" in version.FEATURES

    def test_has_network_config_v2(self):
        assert "NETWORK_CONFIG_V2" in version.FEATURES


class TestVersionString:
    @mock.patch(
        "cloudinit.version._DOWNSTREAM_VERSION", "17.2-3-gb05b9972-0ubuntu1"
    )
    def test_package_version_respected(self):
        """If _DOWNSTREAM_VERSION is filled in, then it should be returned."""
        assert "17.2-3-gb05b9972-0ubuntu1" == version.version_string()

    @mock.patch(
        "cloudinit.version._DOWNSTREAM_VERSION", "@DOWNSTREAM_VERSION@"
    )
    @mock.patch("cloudinit.version.__VERSION__", "17.2")
    def test_package_version_skipped(self):
        """If _PACKAGED_VERSION is not modified, then return __VERSION__."""
        assert "17.2" == version.version_string()
